@@ -8,7 +8,9 @@ use {
         errors::ProgramError,
         file_sum, git,
         launchable::Launchable,
+        path::closest_dir,
         skin::*,
+        stage::Stage,
         task_sync::{Dam, Either},
         verb::Internal,
     },
@@ -21,13 +23,12 @@ use {
     std::{
         io::Write,
         path::PathBuf,
+        sync::{Arc, Mutex},
     },
     strict::NonEmptyVec,
     termimad::{Event, EventSource},
 };
 
-#[cfg(feature = "client-server")]
-use std::sync::{Arc, Mutex};
 
 /// The GUI
 pub struct App {
@@ -54,9 +55,8 @@ pub struct App {
 
     stage_panel: Option<PanelId>,
 
-    /// the root of the active panel
-    #[cfg(feature = "client-server")]
-    root: Arc<Mutex<PathBuf>>,
+    /// an optional copy of the root for the --server
+    shared_root: Option<Arc<Mutex<PathBuf>>>,
 
     /// sender to the sequence channel
     tx_seqs: Sender<Sequence>,
@@ -96,9 +96,7 @@ impl App {
             created_panels_count: 1,
             preview_panel: None,
             stage_panel: None,
-
-            #[cfg(feature = "client-server")]
-            root: Arc::new(Mutex::new(con.launch_args.root.clone())),
+            shared_root: None,
             tx_seqs,
             rx_seqs,
         })
@@ -235,7 +233,6 @@ impl App {
 
     /// if there are exactly two non preview panels, return the selection
     /// in the non focused panel
-    /// FIXME exclude stage panel
     fn get_other_panel_path(&self) -> Option<PathBuf> {
         let len = self.panels.len().get();
         if len == 3 {
@@ -270,7 +267,6 @@ impl App {
         let mut error: Option<String> = None;
         let is_input_invocation = cmd.is_verb_invocated_from_input();
         let app_cmd_context = AppCmdContext {
-            other_path: self.get_other_panel_path(),
             panel_skin,
             preview_panel: self.preview_panel,
             stage_panel: self.stage_panel,
@@ -324,7 +320,6 @@ impl App {
                         let new_input = self.panel().get_input_content();
                         let cmd = Command::from_raw(new_input, false);
                         let app_cmd_context = AppCmdContext {
-                            other_path: self.get_other_panel_path(),
                             panel_skin,
                             preview_panel: self.preview_panel,
                             stage_panel: self.stage_panel,
@@ -344,47 +339,94 @@ impl App {
                 self.tx_seqs.send(sequence).unwrap();
             }
             HandleInApp(internal) => {
-                let new_active_panel_idx = match internal {
-                    Internal::panel_left => {
-                        // we're not here to create a panel, this is handled in the state
-                        // we're here because the state wants us to either move to the panel
-                        // to the left, or close the rightest one
-                        if self.active_panel_idx == 0 {
-                            self.close_panel(self.panels.len().get()-1);
-                            None
-                        } else {
-                            Some(self.active_panel_idx - 1)
+                match internal {
+                    Internal::panel_left | Internal::panel_right => {
+                        let new_active_panel_idx = if internal == Internal::panel_left {
+                            // we're not here to create a panel, this is handled in the state
+                            // we're here because the state wants us to either move to the panel
+                            // to the left, or close the rightest one
+                            if self.active_panel_idx == 0 {
+                                self.close_panel(self.panels.len().get()-1);
+                                None
+                            } else {
+                                Some(self.active_panel_idx - 1)
+                            }
+                        } else { // panel_right
+                            // we're not here to create panels (it's done in the state).
+                            // So we either move to the right or close the leftes panel
+                            if self.active_panel_idx + 1 == self.panels.len().get() {
+                                self.close_panel(0);
+                                None
+                            } else {
+                                Some(self.active_panel_idx + 1)
+                            }
+                        };
+                        if let Some(idx) = new_active_panel_idx {
+                            if is_input_invocation {
+                                self.mut_panel().clear_input();
+                            }
+                            self.active_panel_idx = idx;
+                            let app_cmd_context = AppCmdContext {
+                                panel_skin,
+                                preview_panel: self.preview_panel,
+                                stage_panel: self.stage_panel,
+                                screen: self.screen,
+                                con,
+                            };
+                            self.mut_panel().refresh_input_status(app_state, &app_cmd_context);
                         }
                     }
-                    Internal::panel_right => {
-                        // we're not here to create panels (it's done in the state).
-                        // So we either move to the right or close the leftes panel
-                        if self.active_panel_idx + 1 == self.panels.len().get() {
-                            self.close_panel(0);
-                            None
+                    Internal::toggle_second_tree => {
+                        let panels_count = self.panels.len().get();
+                        let trees_count = self.panels.iter()
+                            .filter(|p| p.state().get_type() == PanelStateType::Tree)
+                            .count();
+                        if trees_count < 2 {
+                            // we open a tree, closing a (non tree) panel if necessary
+                            if panels_count >= con.max_panels_count {
+                                for i in (0..panels_count).rev() {
+                                    if self.panels[i].state().get_type() != PanelStateType::Tree {
+                                        self.close_panel(i);
+                                        break;
+                                    }
+                                }
+                            }
+                            if let Some(selected_path) = self.state().selected_path() {
+                                let dir = closest_dir(selected_path);
+                                if let Ok(Some(new_state)) = BrowserState::new(
+                                    dir,
+                                    self.state().tree_options().without_pattern(),
+                                    self.screen,
+                                    con,
+                                    &Dam::unlimited(),
+                                ) {
+                                    if let Err(s) = self.new_panel(
+                                        Box::new(new_state),
+                                        PanelPurpose::None,
+                                        HDir::Right,
+                                        is_input_invocation,
+                                        con,
+                                    ) {
+                                        error = Some(s);
+                                    }
+                                }
+                            }
                         } else {
-                            Some(self.active_panel_idx + 1)
+                            // we close the rightest inactive tree
+                            for i in (0..panels_count).rev() {
+                                if self.panels[i].state().get_type() == PanelStateType::Tree {
+                                    if i == self.active_panel_idx {
+                                        continue;
+                                    }
+                                    self.close_panel(i);
+                                    break;
+                                }
+                            }
                         }
                     }
                     _ => {
-                        debug!("unhandled propagated internal. cmd={:?}", &cmd);
-                        None
+                        info!("unhandled propagated internal. cmd={:?}", &cmd);
                     }
-                };
-                if let Some(idx) = new_active_panel_idx {
-                    if is_input_invocation {
-                        self.mut_panel().clear_input();
-                    }
-                    self.active_panel_idx = idx;
-                    let app_cmd_context = AppCmdContext {
-                        other_path: self.get_other_panel_path(),
-                        panel_skin,
-                        preview_panel: self.preview_panel,
-                        stage_panel: self.stage_panel,
-                        screen: self.screen,
-                        con,
-                    };
-                    self.mut_panel().refresh_input_status(app_state, &app_cmd_context);
                 }
             }
             Keep => {
@@ -427,7 +469,6 @@ impl App {
                 }
                 if self.remove_state() {
                     let app_cmd_context = AppCmdContext {
-                        other_path: self.get_other_panel_path(),
                         panel_skin,
                         preview_panel: self.preview_panel,
                         stage_panel: self.stage_panel,
@@ -458,13 +499,21 @@ impl App {
         if let Some(text) = error {
             self.mut_panel().set_error(text);
         }
-        self.update_preview(con);
 
-        #[cfg(feature="client-server")]
-        if let Ok(mut root) = self.root.lock() { // when does this not work ?
-            *root = self.state().selected_path().to_path_buf();
+        app_state.other_panel_path = self.get_other_panel_path();
+        if let Some(path) = self.state().tree_root() {
+            app_state.root = path.to_path_buf();
         }
 
+
+        if let Some(shared_root) = &mut self.shared_root {
+            if let Ok(mut root) = shared_root.lock() {
+                *root = app_state.root.clone();
+            }
+        }
+
+        self.update_preview(con);
+
         Ok(())
     }
 
@@ -494,7 +543,7 @@ impl App {
         state: Box<dyn PanelState>,
         purpose: PanelPurpose,
         direction: HDir,
-        is_input_invocation: bool,
+        is_input_invocation: bool, // if true we clean the input
         con: &AppContext,
     ) -> Result<(), String> {
         match state.get_type() {
@@ -558,10 +607,9 @@ impl App {
         con: &AppContext,
     ) -> Result<(), ProgramError> {
         while self.has_pending_task() && !dam.has_event() {
-            if self.do_pending_task(con, dam) {
+            if self.do_pending_task(app_state, con, dam) {
                 self.update_preview(con); // the selection may have changed
                 let app_cmd_context = AppCmdContext {
-                    other_path: self.get_other_panel_path(),
                     panel_skin: &skin.focused,
                     preview_panel: self.preview_panel,
                     stage_panel: self.stage_panel,
@@ -581,13 +629,14 @@ impl App {
     /// do the next pending task
     fn do_pending_task(
         &mut self,
+        app_state: &AppState,
         con: &AppContext,
         dam: &mut Dam,
     ) -> bool {
         let screen = self.screen;
         // we start with the focused panel
         if self.panel().has_pending_task() {
-            self.mut_panel().do_pending_task(screen, con, dam);
+            self.mut_panel().do_pending_task(&app_state.stage, screen, con, dam);
             return true;
         }
         // then the other ones
@@ -595,7 +644,7 @@ impl App {
             if idx != self.active_panel_idx {
                 let panel = &mut self.panels[idx];
                 if panel.has_pending_task() {
-                    panel.do_pending_task(screen, con, dam);
+                    panel.do_pending_task(&app_state.stage, screen, con, dam);
                     return true;
                 }
             }
@@ -619,8 +668,12 @@ impl App {
         let event_source = EventSource::new()?;
         let rx_events = event_source.receiver();
         let mut dam = Dam::from(rx_events);
-        let skin = AppSkin::new(conf, con.launch_args.no_style);
-        let mut app_state = AppState::default();
+        let skin = AppSkin::new(conf, con.launch_args.color == Some(false));
+        let mut app_state = AppState {
+            stage: Stage::default(),
+            root: con.launch_args.root.clone(),
+            other_panel_path: None,
+        };
 
         self.screen.clear_bottom_right_char(w, &skin.focused)?;
 
@@ -630,19 +683,25 @@ impl App {
                 .unwrap();
         }
 
-        #[cfg(feature="client-server")]
+        #[cfg(unix)]
         let _server = con.launch_args.listen.as_ref()
-            .map(|server_name| crate::net::Server::new(
-                &server_name,
-                self.tx_seqs.clone(),
-                Arc::clone(&self.root),
-            ))
+            .map(|server_name| {
+                let shared_root = Arc::new(Mutex::new(app_state.root.clone()));
+                let server = crate::net::Server::new(
+                    &server_name,
+                    self.tx_seqs.clone(),
+                    Arc::clone(&shared_root),
+                );
+                self.shared_root = Some(shared_root);
+                server
+            })
             .transpose()?;
 
         loop {
             if !self.quitting {
                 self.display_panels(w, &skin, &app_state, con)?;
                 time!(
+                    Info,
                     "pending_tasks",
                     self.do_pending_tasks(w, &skin, &mut dam, &mut app_state, con)?,
                 );
